# -*- coding: utf-8 -*-
#
# Copyright 2017 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import logging
from collections import OrderedDict
from livebridge.controldata.controlfile import ControlFile
from livebridge.controldata.dynamo import DynamoControl
from livebridge.controldata.storage import StorageControl

logger = logging.getLogger(__name__)


class ControlData(object):

    CONTROL_DATA_CLIENTS = {
        "file": ControlFile,
        "dynamodb": DynamoControl,
        "sql": StorageControl,
    }

    def __init__(self, config):
        self.config = config
        self.control_client = None
        self.control_data = {}
        self.new_bridges = []
        self.removed_bridges = []

    def _resolve_auth(self, data):
        for x, bridge in enumerate(data.get("bridges", [])):
            if data.get("auth", {}).get(bridge.get("auth")):
                # add user creds to bridge config
                data["bridges"][x]["auth"] = data["auth"][bridge["auth"]]
            for y, target in enumerate(bridge.get("targets", [])):
                if data.get("auth", {}).get(target.get("auth")):
                    # add user creds to target config
                    data["bridges"][x]["targets"][y]["auth"] = data["auth"][target["auth"]]
                elif target.get("auth"):
                    raise LookupError("Data for user [{}] not found in control file.".format(target.get("auth")))
        return data

    def _remove_doubles(self, control_data):
        bridges = []
        filtered = {"auth": control_data.get("auth", {}), "bridges": []}
        # clear double source->target connections
        for bridge in control_data.get("bridges", []):
            # filter sources
            source = copy.deepcopy(bridge)
            source["targets"] = []
            if source not in bridges:
                bridges.append(source)
                filtered["bridges"].append(copy.deepcopy(source))
            # filter targets
            for target in bridge.get("targets", []):
                index = bridges.index(source)
                if target not in filtered["bridges"][index]["targets"]:
                    filtered["bridges"][index]["targets"].append(target)
                else:
                    logger.info("Filtering double target [{}] from source [{}]".format(target, source))
        return filtered

    async def _set_client(self, path):
        if self.control_client:
            return self.control_client

        # set client for control data
        if self.CONTROL_DATA_CLIENTS.get(path):
            self.control_client = self.CONTROL_DATA_CLIENTS.get(path)()
        else:
            self.control_client = self.CONTROL_DATA_CLIENTS.get("file")()

    async def load(self, path, *, resolve_auth=False):
        control_data = {}
        await self._set_client(path)
        control_data = await self.control_client.load(path)

        # filter duplicates
        control_data = self._remove_doubles(control_data)

        if resolve_auth:
            control_data = self._resolve_auth(control_data)

        # detect changes
        self.new_bridges = [OrderedDict(n) for n in control_data.get("bridges", [])
                            if n not in self.control_data.get("bridges", [])]
        self.removed_bridges = [OrderedDict(r) for r in self.control_data.get("bridges", [])
                                if r not in control_data.get("bridges", [])]
        self.control_data = control_data

    async def save(self, path, data):
        await self._set_client(path)
        return await self.control_client.save(path, data)

    def list_new_bridges(self):
        return self.new_bridges

    def list_removed_bridges(self):
        return self.removed_bridges

    def list_bridges(self):
        return self.control_data.get("bridges", [])

    async def check_control_change(self, control_path):
        return await self.control_client.check_control_change(control_path=control_path)



# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import logging
import json
from datetime import datetime
from dateutil.parser import parse as parse_date
import aiobotocore
from botocore.exceptions import BotoCoreError
from livebridge.storages.base import BaseStorage

logger = logging.getLogger(__name__)


class DynamoClient(BaseStorage):

    # date format for datetime values in DynamoDB
    date_fmt = "%Y-%m-%dT%H:%M:%S+00:00"

    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(DynamoClient, cls).__new__(cls)
            logger.debug("Dynamo client: {}".format(cls._instance))
        return cls._instance

    def __init__(self, **kwargs):
        self.access_key = kwargs.get("access_key") or None
        self.secret_key = kwargs.get("secret_key") or None
        self.region = kwargs.get("region")
        self.endpoint_url = kwargs.get("endpoint_url") or None
        self.table_name = kwargs.get("table_name")
        self.table_schema = {
            "TableName": self.table_name,
            "KeySchema": [
                {"AttributeName": "target_id", "KeyType": "HASH"},
                {"AttributeName": "post_id", "KeyType": "RANGE"}
            ],
            "AttributeDefinitions": [
                {"AttributeName": "updated", "AttributeType": "S"},
                {"AttributeName": "target_id", "AttributeType": "S"},
                {"AttributeName": "post_id", "AttributeType": "S"},
                {"AttributeName": "source_id", "AttributeType": "S"}
            ],
            "ProvisionedThroughput": {"ReadCapacityUnits": 3, "WriteCapacityUnits": 3},
            "GlobalSecondaryIndexes": [
                {
                    "IndexName": "source_id-updated-index",
                    "KeySchema": [
                        {"AttributeName": "source_id", "KeyType": "HASH"},
                        {"AttributeName": "updated", "KeyType": "RANGE"}
                    ],
                    "Projection": {"ProjectionType": "KEYS_ONLY"},
                    "ProvisionedThroughput": {"ReadCapacityUnits": 2, "WriteCapacityUnits": 2},
                }
            ]
        }
        self.control_table_name = kwargs.get("control_table_name")
        self.control_table_schema = {
            "TableName": self.control_table_name,
            "KeySchema": [
                {"AttributeName": "id", "KeyType": "HASH"},
            ],
            "AttributeDefinitions": [
                {"AttributeName": "id", "AttributeType": "S"},
            ],
            "ProvisionedThroughput": {"ReadCapacityUnits": 3, "WriteCapacityUnits": 3},
        }

    def __del__(self):
        if hasattr(self, "db_client") and self.db_client:
            self.db_client.close()

    @property
    async def db(self):
        if hasattr(self, "db_client") and self.db_client:
            return self.db_client
        logger.info("DynamoDB: Connecting to region [{}] with endpoint_url [{}] and table [{}]".format(
            self.region, self.endpoint_url or "-", self.table_name or "-"))
        loop = asyncio.get_event_loop()
        session = aiobotocore.get_session(loop=loop)
        self.db_client = session.create_client('dynamodb',
                                               region_name=self.region,
                                               endpoint_url=self.endpoint_url,
                                               aws_secret_access_key=self.secret_key,
                                               aws_access_key_id=self.access_key)
        # dirty fix for disabling crc32 checks
        self.db_client._endpoint._aio_session._default_headers["Accept-Encoding"] = ""
        return self.db_client

    async def setup(self):
        """Setting up DynamoDB table, if it not exists."""
        try:
            client = await self.db
            response = await client.list_tables()
            created = False
            # create table if not already created.
            if self.table_name not in response["TableNames"]:
                logger.info("Creating DynamoDB table [{}]".format(self.table_name))
                resp = await client.create_table(**self.table_schema)
                if resp.get("ResponseMetadata", {}).get("HTTPStatusCode") == 200:
                    logger.info("DynamoDB table [{}] successfully created!".format(self.table_name))
                    created = True
            # create control table if not already created.
            if self.control_table_name and self.control_table_name not in response["TableNames"]:
                logger.info("Creating DynamoDB control_table [{}]".format(self.control_table_name))
                resp = await client.create_table(**self.control_table_schema)
                if resp.get("ResponseMetadata", {}).get("HTTPStatusCode") == 200:
                    logger.info("DynamoDB control table [{}] successfully created!".format(self.control_table_name))
                    created = True
            return created
        except Exception as exc:
            logger.error("[DB] Error when setting up DynamoDB.")
            logger.error(exc)
        return False

    async def scan(self):
        scan_params = {
            "TableName": self.table_name,
            "Select": "ALL_ATTRIBUTES",
            "Limit": 5
        }
        db = await self.db
        try:
            response = await db.scan(**scan_params)
            if response.get("ResponseMetadata", {}).get("HTTPStatusCode") == 200:
                return response["Items"]
        except BotoCoreError as exc:
            logger.error("[DB] Error when scanning db")
            logger.error(exc)
        return []

    async def insert_post(self, **kwargs):
        params = {
            "TableName": self.table_name,
            "Item": {
                "target_id": {"S": kwargs.get("target_id")},
                "post_id": {"S": str(kwargs.get("post_id"))},
                "source_id": {"S": kwargs.get("source_id")},
                "text": {"S": kwargs.get("text") or " "},
                "sticky": {"N": str(int(kwargs.get("sticky", False)))},
                "created": {"S": datetime.strftime(kwargs.get("created"), self.date_fmt)},
                "updated": {"S": datetime.strftime(kwargs.get("updated"), self.date_fmt)},
            }
        }
        # add doc at target if present
        if kwargs.get("target_doc"):
            params["Item"]["target_doc"] = {"S": json.dumps(kwargs["target_doc"])}
        db = await self.db
        try:
            response = await db.put_item(**params)
            if response.get("ResponseMetadata", {}).get("HTTPStatusCode") == 200:
                logger.info("[DB] Post {} {} was saved!".format(kwargs["source_id"], kwargs["post_id"]))
                return True
        except BotoCoreError as exc:
            logger.error("[DB] Error when saving {}".format(kwargs))
            logger.error(exc)
        return False

    async def get_last_updated(self, source_id):
        params = {
            "TableName": self.table_name,
            "IndexName": "source_id-updated-index",
            "KeyConditionExpression": "source_id = :value",
            "ExpressionAttributeValues": {
                ":value": {"S": str(source_id)},
            },
            "ProjectionExpression": "source_id, updated, post_id",
            "ScanIndexForward": False,
            "Limit": 1,
        }
        db = await self.db
        try:
            response = await db.query(**params)
            if response.get("ResponseMetadata", {}).get("HTTPStatusCode") == 200:
                updated = response["Items"][0]["updated"]["S"] if response["Count"] == 1 else None
                tstamp = parse_date(updated) if updated else None
                return tstamp
        except BotoCoreError as exc:
            logger.error("[DB] Error when querying for last updated item on {}".format(source_id))
            logger.error(exc)
        return None

    async def get_known_posts(self, source_id, post_ids):
        results = []
        try:
            # build ExpressionAttributeValues
            expr_attr = {":value": {"S": str(source_id)}}
            for x, pid in enumerate(post_ids):
                expr_attr[":post_id_{}".format(x)] = {"S": str(pid)}

            # build KeyConditionExpression
            filter_expr = " OR ".join(["post_id= :post_id_{}".format(x) for x in range(len(post_ids))])

            params = {
                "TableName": self.table_name,
                "IndexName": "source_id-updated-index",
                "KeyConditionExpression": "source_id = :value ",
                "ExpressionAttributeValues": expr_attr,
                "ScanIndexForward": False,
                "ProjectionExpression": "post_id",
                "FilterExpression": filter_expr,
            }
            db = await self.db
            response = await db.query(**params)
            if response.get("ResponseMetadata", {}).get("HTTPStatusCode") == 200:
                for item in response.get("Items", []):
                    results.append(item["post_id"]["S"])
        except Exception as exc:
            logger.error("[DB] Error when querying for posts {}".format(post_ids))
            logger.exception(exc)
        return results

    async def get_post(self, target_id, post_id):
        params = {
            "TableName": self.table_name,
            "KeyConditionExpression": "target_id = :value AND post_id= :post_id",
            "ExpressionAttributeValues": {
                ":value": {"S": str(target_id)},
                ":post_id": {"S": str(post_id)}
            },
            "ScanIndexForward": False,
            "ProjectionExpression": "target_id, source_id, updated, post_id, target_doc, sticky",
        }
        db = await self.db
        try:
            response = await db.query(**params)
            if response.get("ResponseMetadata", {}).get("HTTPStatusCode") == 200:
                db_post = response["Items"][0] if response["Count"] >= 1 else None
                if db_post:
                    db_post["sticky"] = db_post.get("sticky", {}).get("N")
                    db_post["updated"] = db_post.get("updated", {}).get("S")
                    db_post["source_id"] = db_post.get("source_id", {}).get("S")
                    db_post["post_id"] = db_post.get("post_id", {}).get("S")
                    db_post["target_id"] = db_post.get("target_id", {}).get("S")
                    db_post["target_doc"] = json.loads(db_post["target_doc"]["S"]) \
                        if db_post.get("target_doc", {}).get("S") else {}
                return db_post
        except BotoCoreError as exc:
            logger.error("[DB] Error when querying for a post [{}] on {}".format(post_id, target_id))
            logger.error(exc)
        return None

    async def update_post(self, **kwargs):
        try:
            target_id = kwargs.get("target_id")
            post_id = kwargs.get("post_id")
            if await self.delete_post(target_id, post_id) is True:
                await self.insert_post(**kwargs)
                logger.info("[DB] Post {} {} was updated!".format(target_id, post_id))
                return True
        except Exception as exc:
            logger.error("[DB] Error when updating for a post [{}] on {}".format(
                kwargs.get("post_id"), kwargs.get("target_id")))
            logger.error(exc)
        return False

    async def delete_post(self, target_id, post_id):
        params = {
            "TableName": self.table_name,
            "Key": {
                "target_id": {"S": str(target_id)},
                "post_id": {"S": str(post_id)},
            },
        }
        db = await self.db
        try:
            response = await db.delete_item(**params)
            if response.get("ResponseMetadata", {}).get("HTTPStatusCode") == 200:
                logger.info("[DB] Post {} {} was deleted!".format(target_id, post_id))
                return True
        except Exception as exc:
            logger.error("[DB] Error when deleting for a post [{}] on {}".format(post_id, target_id))
            logger.error(exc)
        return False

    async def get_control(self):
        params = {
            "TableName": self.control_table_name,
            "KeyConditionExpression": "id= :value",
            "ExpressionAttributeValues": {
                ":value": {"S": "control"}
            },
            "ScanIndexForward": False,
        }
        try:
            db = await self.db
            response = await db.query(**params)
            if response.get("ResponseMetadata", {}).get("HTTPStatusCode") == 200:
                db_post = response["Items"][0] if response["Count"] >= 1 else None
                if db_post:
                    return json.loads(db_post["data"]["S"])
        except BotoCoreError as exc:
            logger.error("[DB] Error when querying for control_data on {}".format(self.control_table_name))
            logger.error(exc)
        return False

    async def save_control(self, data):
        params = {
            "TableName": self.control_table_name,
            "ConditionExpression": "id = :value OR attribute_not_exists(id)",
            "ExpressionAttributeValues": {
                ":value": {"S": "control"},
            },
            "Item": {
                "id": {"S": "control"},
                "data": {"S": json.dumps(data)}
            }
        }
        try:
            db = await self.db
            await db.put_item(**params)
            logger.info("[DB] Control data was saved.")
            return True
        except BotoCoreError as exc:
            logger.error("[DB] Error when saving control_data on {}".format(self.control_table_name))
            logger.error(exc)
        return False



# -*- coding: utf-8 -*-
#
# Copyright 2017 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import uuid
from aiohttp import web

logger = logging.getLogger(__name__)

_tokens = {}


async def auth_middleware(app, handler):
    async def middleware_handler(request):
        try:
            # check auth header
            if not request.path == "/api/v1/session":
                token = request.headers.get("X-Auth-Token")
                if token not in _tokens.values():
                    return web.json_response({"error": "Invalid token."}, status=401)
            # return response
            response = await handler(request)
            return response
        except web.HTTPException as ex:
            raise
    return middleware_handler


def error_overrides(overrides):
    async def middleware(app, handler):
        async def middleware_handler(request):
            try:
                response = await handler(request)
                return response
            except web.HTTPException as ex:
                override = overrides.get(ex.status)
                if override:
                    return await override(request, ex)
        return middleware_handler
    return middleware


class WebApi(object):

    def __init__(self, *, config, controller, loop):
        self.config = config
        self.loop = loop

        # start server
        logger.info("Starting API server ...")
        middlewares = [error_overrides({405: self.handle_405}), auth_middleware]
        self.app = web.Application(loop=loop, middlewares=middlewares)
        self.app["controller"] = controller
        self.app.router.add_get("/", self.index_handler)
        self.app.router.add_get("/api/v1/controldata", self.control_get)
        self.app.router.add_put("/api/v1/controldata", self.control_put)
        self.app.router.add_post("/api/v1/session", self.login, expect_handler=web.Request.json)
        self.handler = self.app.make_handler()
        f = self.loop.create_server(self.handler, self.config["host"], self.config["port"])
        self.srv = loop.run_until_complete(f) if not loop.is_running() else None
        logger.info("... API server started up")

    async def login(self, request):
        try:
            assert self.config["auth"]["user"]
            assert self.config["auth"]["password"]
        except AssertionError:
            logger.error("HTTP Auth credentials are missing!")
            return web.json_response({"error": "Auth credentials are missing."}, status=400)

        params = await request.post()
        user = params.get('username', None)
        if (user == self.config["auth"]["user"] and
                params.get('password', None) == self.config["auth"]["password"]):
            # User is in our database, remember their login details
            _tokens[user] = str(uuid.uuid4())
            return web.json_response({"token": _tokens[user]})
        return web.json_response({"error": "Unauthorized"}, status=401)

    async def handle_405(self, request, response):
        return web.json_response({"error": "Method Not Allowed"}, status=405)

    async def index_handler(self, request):
        return web.json_response({})

    async def control_get(self, request):
        control_data = await self.app["controller"].load_control_data(resolve_auth=False)
        return web.json_response(control_data.control_data)

    async def control_put(self, request):
        try:
            await request.post()
            if request.has_body:
                uploaded_doc = await request.json()
                res = await self.app["controller"].save_control_data(uploaded_doc)
                if res:
                    return web.json_response({"ok": "true"})
                else:
                    return web.json_response({"error": "Controldata was not saved."}, status=400)
            else:
                return web.json_response({"error": "No request body was found."}, status=400)

        except Exception as exc:
            logger.error("Error handling PUT controldata")
            logger.exception(exc)
            return web.json_response({"error": "Internal Server Error"}, status=500)

    def shutdown(self):
        logger.debug("Shutting down web API!")
        if self.srv:
            self.srv.close()
        self.loop.run_until_complete(self.srv.wait_closed())
        self.loop.run_until_complete(self.app.shutdown())
        self.loop.run_until_complete(self.handler.shutdown(60.0))
        self.loop.run_until_complete(self.app.cleanup())



# -*- coding: utf-8 -*-
#
# Copyright 2016, 2017 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import aiobotocore
import json
import logging
import os
import os.path
import yaml
from botocore.exceptions import ClientError
from livebridge.config import AWS
from livebridge.controldata.base import BaseControl

logger = logging.getLogger(__name__)


class ControlFile(BaseControl):

    def __init__(self):
        self._sqs_client = None
        self._s3_client = None
        self.config = AWS
        self._updated_local = None

    def __del__(self):
        if self._sqs_client:
            self._sqs_client.close()

        if self._s3_client:
            self._s3_client.close()

    @property
    async def sqs_client(self):
        if self._sqs_client:
            return self._sqs_client
        session = aiobotocore.get_session()
        self._sqs_client = session.create_client(
            'sqs',
            region_name=self.config["region"],
            aws_secret_access_key=self.config["secret_key"] or None,
            aws_access_key_id=self.config["access_key"] or None)

        await self._purge_sqs_queue()

        return self._sqs_client

    @property
    def s3_client(self):
        if self._s3_client:
            return self._s3_client
        session = aiobotocore.get_session()
        self._s3_client = session.create_client(
            's3',
            region_name=self.config["region"],
            aws_secret_access_key=self.config["secret_key"] or None,
            aws_access_key_id=self.config["access_key"] or None)
        return self._s3_client

    async def _purge_sqs_queue(self):
        # purge queue before starting watching
        try:
            await self._sqs_client.purge_queue(
                QueueUrl=self.config["sqs_s3_queue"]
            )
            logger.info("Purged SQS queue {}".format(self.config["sqs_s3_queue"]))
        except ClientError as exc:
            logger.warning("Purging SQS queue failed with: {}".format(exc))

    async def check_control_change(self, control_path=None):
        if not self.config.get("sqs_s3_queue", False):
            return await self._check_local_changes(control_path)
        else:
            return await self._check_s3_changes()

    async def _check_local_changes(self, control_path):
        is_changed = False
        try:
            last_updated = os.stat(control_path).st_mtime
            if last_updated != self._updated_local:
                is_changed = True
            self._updated_local = last_updated
        except Exception as exc:
            logger.error("Error fetching last updated time of local control file: {}".format(exc))
        return is_changed

    async def _check_s3_changes(self):
        client = await self.sqs_client
        # check for update events
        try:
            response = await client.receive_message(
                QueueUrl=self.config["sqs_s3_queue"]
            )
            for msg in response.get("Messages", []):
                logger.debug("SQS {}".format(msg.get("MessageId")))
                body = msg.get("Body")
                data = json.loads(body) if body else None
                await client.delete_message(
                    QueueUrl=self.config["sqs_s3_queue"],
                    ReceiptHandle=msg.get("ReceiptHandle")
                )
                if data:
                    for rec in data.get("Records", []):
                        logger.debug("EVENT: {} {}".format(
                            rec.get("s3", {}).get("object", {}).get("key"), rec.get("eventName")))
                        return True
        except Exception as exc:
            logger.error("Error fetching SQS messages with: {}".format(exc))
        return False

    async def load(self, path, *, resolve_auth=False):
        if not path.startswith("s3://"):
            body = self._load_from_file(path)
        else:
            body = await self._load_from_s3(path)
        return yaml.load(body)

    def _load_from_file(self, path):
        logger.info("Loading control file from disk: {}".format(path))
        if not os.path.exists(path):
            raise IOError("Path for control file not found.")

        file = open(path, "r")
        body = file.read()
        self._updated_local = os.stat(path).st_mtime
        file.close()

        return body

    async def _save_to_file(self, path, data):
        logger.info("Saving control file to disk.")
        directory = os.path.dirname(os.path.abspath(path))
        if not os.access(directory, os.W_OK):
            raise IOError("Path for control file not writable: {}".format(path))

        file = open(path, "w+")
        file.write(data)
        file.close()

        return True

    async def _load_from_s3(self, url):
        bucket, key = url.split('/', 2)[-1].split('/', 1)
        logger.info("Loading control file from s3: {} - {}".format(bucket, key))
        control_file = await self.s3_client.get_object(Bucket=bucket, Key=key)
        control_data = await control_file["Body"].read()
        return control_data

    async def _save_to_s3(self, path, data):
        bucket, key = path.split('/', 2)[-1].split('/', 1)
        logger.info("Saving control file to s3: {} - {}".format(bucket, key))
        await self.s3_client.put_object(Body=data, Bucket=bucket, Key=key)
        return True

    async def save(self, path, data):
        res = False
        yaml_data = yaml.dump(data, indent=4, default_flow_style=False)
        if not path.startswith("s3://"):
            res = await self._save_to_file(path, yaml_data)
        else:
            res = await self._save_to_s3(path, yaml_data)
        return res



# -*- coding: utf-8 -*-
#
# Copyright 2017 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dsnparse
import logging
from datetime import datetime
from pymongo import DESCENDING
from motor.motor_asyncio import AsyncIOMotorClient
from bson.objectid import ObjectId
from livebridge.storages.base import BaseStorage


logger = logging.getLogger(__name__)


class MongoStorage(BaseStorage):

    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(MongoStorage, cls).__new__(cls)
            logger.debug("MongoDB client: {}".format(cls._instance))
        return cls._instance

    def __init__(self, **kwargs):
        self.dsn = kwargs.get("dsn", None)
        self.table_name = kwargs.get("table_name")
        self.control_table_name = kwargs.get("control_table_name")

        # get db name
        info = dsnparse.parse(self.dsn)
        self.db_name = info.paths.pop() if len(info.paths) == 1 else ""
        assert self.db_name != "", "No database name provided in DSN connection string"

    @property
    async def db(self):
        if hasattr(self, "_db") and self._db:
            return self._db
        logger.debug("Connecting to {}".format(self.dsn))
        client = AsyncIOMotorClient(self.dsn)
        self._db = client[self.db_name]
        return self._db

    async def setup(self):
        """Setting up MongoDB collections, if they not exist."""
        try:
            db = await self.db
            collections = await db.collection_names()
            created = False
            if self.table_name not in collections:
                # create table
                logger.info("Creating MongoDB collection [{}]".format(self.table_name))
                await db.create_collection(self.table_name)
                await db[self.table_name].create_index([("target_id", DESCENDING), ("post_id", DESCENDING)])
                created = True
            # create control collection if not already created.
            if self.control_table_name not in collections:
                # create table
                logger.info("Creating MongoDB control data collection [{}]".format(self.control_table_name))
                await db.create_collection(self.control_table_name)
                created = True
            return created
        except Exception as exc:
            logger.error("[DB] Error when setting up MongoDB collections: {}".format(exc))
        return False

    async def get_last_updated(self, source_id):
        try:
            coll = (await self.db)[self.table_name]
            cursor = coll.find({"source_id": source_id})
            cursor.sort("updated", -1).limit(1)
            async for doc in cursor:
                if doc.get("updated"):
                    return doc["updated"]
        except Exception as exc:
            logger.error("[DB] Error when querying for last updated item on {}".format(source_id))
            logger.exception(exc)
        return None

    async def get_known_posts(self, source_id, post_ids):
        results = []
        try:
            object_ids = list(map(lambda x: ObjectId(x), post_ids))
            coll = (await self.db)[self.table_name]
            cursor = coll.find({"source_id": source_id, "_id": {"$in": object_ids}})
            async for doc in cursor:
                results.append(str(doc["_id"]))
        except Exception as exc:
            logger.error("[DB] Error when querying for posts {}".format(post_ids))
            logger.exception(exc)
        return results

    async def get_post(self, target_id, post_id):
        try:
            coll = (await self.db)[self.table_name]
            doc = await coll.find_one({"target_id": target_id, "post_id": post_id})
            if doc:
                doc["_id"] = str(doc["_id"])
                return doc
        except Exception as exc:
            logger.error("[DB] Error when querying for a post [{}] on {}".format(post_id, target_id))
            logger.error(exc)
        return None

    async def insert_post(self, **kwargs):
        try:
            target_id = kwargs.get("target_id")
            doc = {
                "target_id": target_id,
                "post_id": str(kwargs.get("post_id")),
                "source_id": kwargs.get("source_id"),
                "text": kwargs.get("text") or " ",
                "sticky": str(int(kwargs.get("sticky", False))),
                "created": kwargs.get("created"),
                "updated": kwargs.get("updated"),
                "target_id": target_id,
                "target_doc": kwargs.get("target_doc", "")
            }
            coll = (await self.db)[self.table_name]
            await coll.insert_one(doc)
            logger.info("[DB] Post {} {} was saved!".format(kwargs["source_id"], kwargs["post_id"]))
            return True
        except Exception as exc:
            logger.error("[DB] Error when saving {}".format(kwargs))
            logger.error(exc)
        return False

    async def update_post(self, **kwargs):
        try:
            target_id = kwargs.get("target_id")
            doc = {
                "target_id": target_id,
                "post_id": str(kwargs.get("post_id")),
                "source_id": kwargs.get("source_id"),
                "text": kwargs.get("text") or " ",
                "sticky": str(int(kwargs.get("sticky", 0))),
                "created": kwargs.get("created"),
                "updated": kwargs.get("updated"),
                "target_id": target_id,
                "target_doc": kwargs.get("target_doc", "")
            }
            coll = (await self.db)[self.table_name]
            await coll.replace_one({"target_id": kwargs.get("target_id"), "post_id": kwargs.get("post_id")}, doc)
            logger.info("[DB] Post {} {} was updated!".format(kwargs.get("post_id"), kwargs.get("target_id")))
            return True
        except Exception as exc:
            logger.error("[DB] Error when updating for a post [{}] on {}".format(
                kwargs.get("post_id"), kwargs.get("target_id")))
            logger.error(exc)
        return False

    async def delete_post(self, target_id, post_id):
        try:
            coll = (await self.db)[self.table_name]
            await coll.remove({"target_id": target_id, "post_id": post_id}, {"justOne": True})
            logger.info("[DB] Post {} {} was deleted!".format(target_id, post_id))
            return True
        except Exception as exc:
            logger.error("[DB] Error when deleting for a post [{}] on {}".format(post_id, target_id))
            logger.error(exc)
        return False

    async def get_control(self, updated=None):
        try:
            query = {"type": "control"}
            if updated:
                query["updated"] = {"$gt": updated}
            coll = (await self.db)[self.control_table_name]
            doc = await coll.find_one(query)
            if doc:
                return doc
        except Exception as exc:
            logger.error("[DB] Error when querying for a control data on {}".format(self.control_table_name))
            logger.error(exc)
        return False

    async def save_control(self, data):
        try:
            query = {"type": "control"}
            doc = {"type": "control", "data": data, "updated": datetime.now()}
            coll = (await self.db)[self.control_table_name]
            res = await coll.replace_one(query, doc, upsert=True)
            if res.modified_count != 1 and not res.upserted_id:
                logger.error("[DB] Control data was not saved.")
            else:
                logger.info("[DB] Control data was saved.")
                return True
        except Exception as exc:
            logger.error("[DB] Error when saving control data on {}".format(self.control_table_name))
            logger.error(exc)
        return False



# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pkgutil
import inspect
import logging
from importlib import import_module
from livebridge.base import BaseTarget, BasePost, BaseSource, BaseConverter
from livebridge.components import add_converter, add_post, add_target, add_source

logger = logging.getLogger(__name__)


def load_extensions():
    for pkg_info in pkgutil.walk_packages(onerror=importfail):
        if pkg_info[1].startswith("livebridge"):
            if pkg_info[2] is True:
                get_extensions(pkg_info)


def importfail(pkg):
    """Just do nothing when module-walk fails."""
    pass


def get_extensions(pkg):
    mod = import_module(pkg[1])
    for member in [d for d in inspect.getmembers(mod) if not d[0].startswith("__") and inspect.isclass(d[1])]:
        ext = member[1]
        if issubclass(ext, BaseTarget):
            logger.debug("Loading target: {}".format(ext))
            add_target(ext)
        elif issubclass(ext, BaseSource):
            logger.debug("Loading source: {}".format(ext))
            add_source(ext)
        elif issubclass(ext, BasePost):
            logger.debug("Loading post: {}".format(ext))
            add_post(ext)
        elif issubclass(ext, BaseConverter):
            logger.debug("Loading converter: {}".format(ext))
            add_converter(ext)



# -*- coding: utf-8 -*-
#
# Copyright 2016, 2017 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

LOGLEVEL = os.environ.get("LB_LOGLEVEL", "INFO")

LOGFILE = os.environ.get("LB_LOGFILE")

CONTROLFILE = os.environ.get("LB_CONTROLFILE")

POLL_INTERVAL = int(os.environ.get("LB_POLL_INTERVAL", 60))
POLL_CONTROL_INTERVAL = int(os.environ.get("LB_POLL_CONTROL_INTERVAL", 60))

DB = {
    "dsn": os.environ.get("LB_DB_DSN"),
    "table_name": os.environ.get("LB_DB_TABLE", "livebridge_dev"),
    "control_table_name": os.environ.get("LB_DB_CONTROL_TABLE")
}

AWS = {
    "access_key": os.environ.get("LB_AWS_ACCESS_KEY", ""),
    "secret_key": os.environ.get("LB_AWS_SECRET_KEY", ""),
    "region": os.environ.get("LB_AWS_REGION", "eu-central-1"),
    "endpoint_url": os.environ.get("LB_DYNAMO_ENDPOINT"),
    "table_name": os.environ.get("LB_DYNAMO_TABLE", "livebridge-dev"),
    "control_table_name": os.environ.get("LB_DYNAMO_CONTROL_TABLE"),
    "sqs_s3_queue": os.environ.get("LB_SQS_S3_QUEUE", ""),
}

WEB = {
    "host": os.environ.get("LB_WEB_HOST"),
    "port": os.environ.get("LB_WEB_PORT"),
    "auth": {
        "user": os.environ.get("LB_WEB_USER"),
        "password": os.environ.get("LB_WEB_PWD")
    }
}



# -*- coding: utf-8 -*-
#
# Copyright 2016, 2017 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import copy
import logging
from livebridge.components import get_source, get_db_client, get_hash
from livebridge.base import InvalidTargetResource

logger = logging.getLogger(__name__)


class LiveBridge(object):

    def __init__(self, config):
        self.targets = []
        self.api_client = None
        self.config = config
        self.hash = get_hash(self.config)
        self.source_id = self.config.get("source_id")
        self.endpoint = self.config.get("endpoint")
        self.label = self.config.get("label", "-")
        self.db = get_db_client()
        self.queue = asyncio.Queue()
        self.queue_task = None
        self.sleep_tasks = []

    def __repr__(self):
        return "<LiveBridge [{}] {} {}>".format(self.label, self.endpoint, self.source_id)

    @property
    def source(self):
        if self.api_client:
            return self.api_client
        # setup api client
        self.api_client = get_source(self.config)
        return self.api_client

    def stop(self):
        # stop queue task first
        if self.queue_task:
            self.queue_task.cancel()
        # stop sleeping tasks
        for task in self.sleep_tasks:
            task.cancel()
        return True

    def add_target(self, target):
        logger.debug("Adding target {} to {}".format(target, self))
        self.targets.append(target)

    async def listen_ws(self):
        return asyncio.Task(self.source.listen(self.new_posts))

    async def check_posts(self):
        try:
            posts = await self.source.poll()
            if posts:
                asyncio.ensure_future(self.new_posts(posts))
        except Exception as exc:
            logger.error("Fatal checking {}{}".format(getattr(self, "endpoint", "-"), self.source_id))
            logger.exception(exc)
        return True

    async def _sleep(self, seconds):
        try:
            task = asyncio.ensure_future(asyncio.sleep(seconds))
            self.sleep_tasks.append(task)
            await task
        except asyncio.CancelledError:
            pass
        finally:
            self.sleep_tasks.remove(task)
        return True

    async def _action_done(self, future, item):
        exc = future.exception()
        if exc and type(exc) is InvalidTargetResource:
            logger.warning("POST {post.id} not distributed to {target.target_id} [{count}], no retry.".format(**item))
            logger.warning(exc)
        elif exc:
            logger.error("TARGET ACTION FAILED, WILL RETRY: [{}] {} {} [{}]".format(
                item["count"], item["post"], item["target"], exc))
            item["count"] = item["count"] + 1
            await self._sleep(5 * item["count"])
            await self._put_to_queue(item)
        else:
            logger.info("POST {post.id} distributed to {target.target_id} [{count}]".format(**item))

    async def _put_to_queue(self, item):
        if not self.queue_task:
            self.queue_task = asyncio.ensure_future(self._queue_consumer())
        await self.queue.put(item)

    async def _process_action(self, task):
        fut = asyncio.ensure_future(task["target"].handle_post(task["post"]))
        callback = lambda fn: asyncio.ensure_future(self._action_done(fn, task))
        fut.add_done_callback(callback)

    async def _queue_consumer(self):
        try:
            while True:
                task = await self.queue.get()
                self.queue.task_done()
                if task["count"] > 10:
                    logger.debug("DISTRIBUTION ABORTED: {post.id} {target.target_id} [{count}]".format(**task))
                    continue
                asyncio.ensure_future(self._process_action(task))
        except asyncio.CancelledError:
            pass
        except Exception as exc:
            logger.error("WORKER QUEUE FAILED: {}".format(exc))
            logger.exception(exc)

    async def new_posts(self, posts):
        try:
            logger.info("##### Received {} posts from {}".format(len(posts), self.source))
            for new_post in posts:
                for target in self.targets:
                    post = copy.deepcopy(new_post)
                    item = {"post": post, "target": target, "count": 0}
                    await self._put_to_queue(item)
        except Exception as exc:
            logger.error("Handling new posts failed [{}]: {}".format(self.source, exc))



# -*- coding: utf-8 -*-
#
# Copyright 2017 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from livebridge.storages.base import BaseStorage


logger = logging.getLogger(__name__)


class DummyStorage(BaseStorage):

    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(DummyStorage, cls).__new__(cls)
            logger.debug("Dummy client: {}".format(cls._instance))
        return cls._instance

    def __init__(self, **kwargs):
        pass

    async def setup(self):
        return False

    async def get_last_updated(self, source_id):
        return None

    async def get_known_posts(self, source_id, post_ids):
        return []

    async def get_post(self, target_id, post_id):
        return None

    async def insert_post(self, **kwargs):
        return True

    async def update_post(self, **kwargs):
        return True

    async def delete_post(self, target_id, post_id):
        return True

    async def get_control(self, updated=None):
        return False
